# Test for sparkmeasure/jmx.py
import re

# Note this requires pytest and pyspark to be installed
from . import StageMetrics
from .jmx import jmxexport
from .testutils import setup_sparksession

def test_jmxexport(setup_sparksession):
    spark = setup_sparksession
    conf = spark.sparkContext.getConf()
    conf.set("spark.sql.shuffle.partitions", "2")
    conf.set("spark.default.parallelism", "2")
    stagemetrics = StageMetrics(spark)

    stagemetrics.begin()
    spark.sql("select count(*) from range(1000) cross join range(1000) cross join range(1000)").show()
    stagemetrics.end()

    # print report to standard output
    stagemetrics.print_report()

    # get metrics data as a dictionary
    metrics = stagemetrics.aggregate_stagemetrics()
    print(f"metrics elapsedTime = {metrics.get('elapsedTime')}")
    assert metrics.get('elapsedTime') > 0

    jmxexport(spark, metrics)

    dropwizard = spark._jvm.ch.cern.metrics.DropwizardMetrics

    # Regex pour matcher la sortie attendue avec une étoile pour resultSize
    expected_regex = (
        r"Metrics for namespace 'unknown' and pod 'unknown'\n"
        r"Total gauges: 2\n"
        r"Gauge: unknown\.unknown\.resultSize = [0-9]+(\.[0-9]+)?\n"
        r"Gauge: unknown\.unknown\.peakExecutionMemory = 0\.0\n"
        r"Total counters: 19\n"
        r"Counter: unknown\.unknown\.bytesRead = 0\n"
        r"Counter: unknown\.unknown\.numStages = 3\n"
        r"Counter: unknown\.unknown\.shuffleRecordsWritten = [0-9]+\n"
        r"Counter: unknown\.unknown\.shuffleRemoteBytesRead = 0\n"
        r"Counter: unknown\.unknown\.shuffleLocalBlocksFetched = [0-9]+\n"
        r"Counter: unknown\.unknown\.shuffleTotalBlocksFetched = [0-9]+\n"
        r"Counter: unknown\.unknown\.memoryBytesSpilled = 0\n"
        r"Counter: unknown\.unknown\.bytesWritten = 0\n"
        r"Counter: unknown\.unknown\.numTasks = [0-9]+\n"
        r"Counter: unknown\.unknown\.recordsWritten = 0\n"
        r"Counter: unknown\.unknown\.shuffleRecordsRead = [0-9]+\n"
        r"Counter: unknown\.unknown\.recordsRead = 2000\n"
        r"Counter: unknown\.unknown\.shuffleLocalBytesRead = [0-9]+\n"
        r"Counter: unknown\.unknown\.shuffleBytesWritten = [0-9]+\n"
        r"Counter: unknown\.unknown\.shuffleTotalBytesRead = [0-9]+\n"
        r"Counter: unknown\.unknown\.metrics_published_total = 30\n"
        r"Counter: unknown\.unknown\.diskBytesSpilled = 0\n"
        r"Counter: unknown\.unknown\.shuffleRemoteBytesReadToDisk = 0\n"
        r"Counter: unknown\.unknown\.shuffleRemoteBlocksFetched = 0\n"
        r"Total timers: 10\n"
        r"Timer: unknown.unknown.shuffleWriteTime = count=1\n"
        r"Timer: unknown.unknown.stageDuration = count=1\n"
        r"Timer: unknown.unknown.executorCpuTime = count=1\n"
        r"Timer: unknown.unknown.shuffleFetchWaitTime = count=1\n"
        r"Timer: unknown.unknown.executorRunTime = count=1\n"
        r"Timer: unknown.unknown.jvmGCTime = count=1\n"
        r"Timer: unknown.unknown.elapsedTime = count=1\n"
        r"Timer: unknown.unknown.executorDeserializeCpuTime = count=1\n"
        r"Timer: unknown.unknown.resultSerializationTime = count=1\n"
        r"Timer: unknown.unknown.executorDeserializeTime = count=1\n"
    )

    output = dropwizard.describeMetrics()
    print(output)
    assert re.search(expected_regex, output, re.MULTILINE | re.DOTALL), f"Output does not match expected regex.\nGot:\n{output}\nWant:\n{expected_regex}"

    spark.stop()



# Test for sparkmeasure/stagemetrics.py
from . import StageMetrics
from .testutils import setup_sparksession

def test_stagemetrics(setup_sparksession):
    spark = setup_sparksession
    stagemetrics = StageMetrics(spark)

    stagemetrics.begin()
    spark.sql("select count(*) from range(1000) cross join range(1000) cross join range(1000)").show()
    stagemetrics.end()

    # print report to standard output
    stagemetrics.print_report()

    # get metrics data as a dictionary
    metrics = stagemetrics.aggregate_stagemetrics()
    print(f"metrics elapsedTime = {metrics.get('elapsedTime')}")
    assert metrics.get('elapsedTime') > 0

    # map metrics into Spark DataFrames
    df = stagemetrics.create_stagemetrics_DF("PerfStageMetrics")
    aggregatedDF = stagemetrics.aggregate_stagemetrics_DF("PerfStageMetrics")

    assert df.count() > 0
    assert aggregatedDF.count() > 0

    spark.stop()



"""sparkmeasure is a tool for performance troubleshooting of Apache Spark workloads.
It simplifies the collection and analysis of Spark task metrics data.
See details at https://github.com/LucaCanali/sparkMeasure
"""


class TaskMetrics:
    """TaskMetrics class provides the API to collect and process task metrics data aggregated by task execution
       This is a finer granularity than StageMetrics and potentially collects much more data.
       This is a Python wrapper class to the corresponding Scala class of sparkMeasure."""
    def __init__(self, sparksession):
        self.sparksession = sparksession
        self.sc = self.sparksession.sparkContext
        try:
            self.taskmetrics = self.sc._jvm.ch.cern.sparkmeasure.TaskMetrics(self.sparksession._jsparkSession)
        except Exception as e:
            # Handle the case where the sparkmeasure jar is not found
            print("The class ch.cern.sparkmeasure.TaskMetrics does not exist or could not be loaded.", e)
            print("\nError: the sparkMeasure jar is like not loaded")
            print("Please check configuration: spark.jars.packages, spark.jars and/or driver classpath configuration and re-run the workload")

    def begin(self):
        """Begin collecting task metrics data."""
        self.taskmetrics.begin()

    def end(self):
        """End collecting task metrics data."""
        self.taskmetrics.end()

    def report(self):
        """Return a report of the collected task metrics data."""
        return self.taskmetrics.report()

    def aggregate_taskmetrics(self):
        """Return a dictionary of the aggregated task metrics data."""
        return self.taskmetrics.aggregateTaskMetricsJavaMap()

    def print_report(self):
        """Print a report of the collected task metrics data."""
        print(self.report())

    def runandmeasure(self, env, codetorun):
        """Run a Python code snippet and collect task metrics data."""
        self.begin()
        exec(codetorun, env)
        self.end()
        self.print_report()

    def create_taskmetrics_DF(self, viewname="PerfTaskMetrics"):
        """Create a Spark Dataframe from the collected task metrics data."""
        df = self.taskmetrics.createTaskMetricsDF(viewname)
        # convert the returned Java object to a Python Dataframe
        from pyspark.sql.dataframe import DataFrame
        return DataFrame(df, self.sparksession)

    def aggregate_taskmetrics_DF(self, viewname="PerfTaskMetrics"):
        """Create a Spark Dataframe from the aggregated task metrics data."""
        df = self.taskmetrics.aggregateTaskMetrics(viewname)
        # convert the returned Java object to a Python Dataframe
        from pyspark.sql.dataframe import DataFrame
        return DataFrame(df, self.sparksession)

    def save_data(self, df, filepathandname, fileformat):
        """Save the collected task metrics data to a file."""
        df.repartition(1).write.format(fileformat).save(filepathandname)

    def remove_listener(self):
        """Remove the listener from the SparkContext."""
        self.taskmetrics.removeListener()



# Note this requires pytest and pyspark to be installed
import pytest
from pyspark.sql import SparkSession
import glob
import os

@pytest.fixture
def setup_sparksession():
    # Note this is supposed to run after sbt package on the sparkmeasure project
    # so that we can get the jar file from the target folder
    SPARKMEASURE_JAR_FOLDER = "target/scala-2.12/"

    # Scan SPARKMEASURE_JAR_FOLDER for the spark-measure jar file
    # Set the directory you want to scan
    pattern = 'spark-measure_*.jar'  # Change this to your desired pattern

    # Use glob to find files matching the pattern
    files = glob.glob(os.path.join(SPARKMEASURE_JAR_FOLDER, pattern))

    if len(files) == 0:
        raise FileNotFoundError(f"No files matching pattern '{pattern}' found in {SPARKMEASURE_JAR_FOLDER}")
    elif len(files) > 1:
        raise FileExistsError(f"Multiple files matching pattern '{pattern}' found in {SPARKMEASURE_JAR_FOLDER}: {files}")

    jarfile=files[0]

    spark = (SparkSession.builder
            .appName("Test sparkmeasure instrumentation of Python/PySpark code")
            .master("local[*]")
            .config("spark.jars", jarfile)
            .getOrCreate()
            )
    return spark


# Test for sparkmeasure/stagemetrics.py
from . import TaskMetrics
from .testutils import setup_sparksession


def test_stagemetrics(setup_sparksession):
    spark = setup_sparksession
    taskmetrics = TaskMetrics(spark)

    taskmetrics.begin()
    spark.sql("select count(*) from range(1000) cross join range(1000) cross join range(1000)").show()
    taskmetrics.end()

    # print report to standard output
    taskmetrics.print_report()

    # get metrics data as a dictionary
    metrics = taskmetrics.aggregate_taskmetrics()
    print(f"metrics elapsedTime = {metrics.get('elapsedTime')}")
    assert metrics.get('numTasks') > 0

    # map metrics into Spark DataFrames
    df = taskmetrics.create_taskmetrics_DF("PerfTaskMetrics")
    aggregatedDF = taskmetrics.aggregate_taskmetrics_DF("PerfTaskMetrics")

    assert df.count() > 0
    assert aggregatedDF.count() > 0

    spark.stop()



"""sparkmeasure is a tool for performance troubleshooting of Apache Spark workloads.
It simplifies the collection and analysis of Spark task metrics data.
See details at https://github.com/LucaCanali/sparkMeasure
"""


class StageMetrics:
    """StageMetrics class provides the API to collect and process task metrics data aggregated by execution stage.
       This is a Python wrapper class to the corresponding Scala class of sparkMeasure."""
    def __init__(self, sparksession):
        self.sparksession = sparksession
        self.sc = self.sparksession.sparkContext
        try:
            self.stagemetrics = self.sc._jvm.ch.cern.sparkmeasure.StageMetrics(self.sparksession._jsparkSession)
        except Exception as e:
            # Handle the case where the sparkmeasure jar is not found
            print("The class ch.cern.sparkmeasure.StageMetrics does not exist or could not be loaded.", e)
            print("\nError: the sparkMeasure jar is like not loaded")
            print("Please check configuration: spark.jars.packages, spark.jars and/or driver classpath configuration and re-run the workload")

    def begin(self):
        """Begin collecting stage metrics data."""
        self.stagemetrics.begin()

    def end(self):
        """End collecting stage metrics data."""
        self.stagemetrics.end()

    def report(self):
        """Return a report of the collected stage metrics data."""
        return self.stagemetrics.report()

    def aggregate_stagemetrics(self):
        """Return a dictionary of the aggregated stage metrics data."""
        return self.stagemetrics.aggregateStageMetricsJavaMap()

    def print_report(self):
        """Print a report of the collected stage metrics data."""
        print(self.report())

    def report_memory(self):
        """Return the collected stage metrics data for memory usage."""
        return self.stagemetrics.reportMemory()

    def print_memory_report(self):
        """Print a report of the collected stage metrics data for memory usage."""
        print(self.report_memory())

    def runandmeasure(self, env, codetorun):
        """Run a Python code snippet and collect stage metrics data."""
        self.begin()
        exec(codetorun, env)
        self.end()
        self.print_report()

    def create_stagemetrics_DF(self, viewname="PerfStageMetrics"):
        """Create a Spark Dataframe from the collected stage metrics data."""
        df = self.stagemetrics.createStageMetricsDF(viewname)
        # convert the returned Java object to a Python Dataframe
        from pyspark.sql.dataframe import DataFrame
        return DataFrame(df, self.sparksession)

    def aggregate_stagemetrics_DF(self, viewname="PerfStageMetrics"):
        """Create a Spark Dataframe from the aggregated stage metrics data."""
        df = self.stagemetrics.aggregateStageMetrics(viewname)
        # convert the returned Java object to a Python Dataframe
        from pyspark.sql.dataframe import DataFrame
        return DataFrame(df, self.sparksession)

    def save_data(self, df, filepathandname, fileformat="json"):
        """Save the collected stage metrics data to a file."""
        df.repartition(1).write.format(fileformat).save(filepathandname)

    def remove_listener(self):
        """Remove the Spark listener that collects stage metrics data."""
        self.stagemetrics.removeListener()



from typing import Union, Dict

import logging
from sparkmeasure import StageMetrics
import datetime

logger = logging.getLogger(__name__)

def jmxexport(spark_session, metrics: Dict[str, Union[float, int]]):
    """Publish metrics to Dropwizard via JMX."""
    try:
        publish_metrics_count = 0
        dropwizard = spark_session._jvm.ch.cern.metrics.DropwizardMetrics
        for key, value in metrics.items():
            # or setGauge
            is_counter = True
            dropwizard.setMetricAutoType(key, float(value))
            # Example counter to track the number of times metrics have been published
            publish_metrics_count += 1

        dropwizard.setMetricAutoType("metrics_published_total", float(publish_metrics_count))

        logger.info("%d Dropwizard metrics published via JMX", len(metrics))
    except Exception as e:
        logger.error("Failed to publish Dropwizard metrics: %s", e)
